/**
 * ODIN overlay management middleware toolkit
 * 
 * ODIN Group, College of Computing, Georgia Institute of Technology
 *
 * For more information:
 * http://www.cc.gatech.edu/~cooperb/odin/
 * odin.gatech.edu
 *
 * You are free to use, modify and redistribute this software. 
 * However, we make no warranty for this software, and by downloading it
 * you agree not to hold use liable for anything. Do not under any
 * circumstances use this software to share copyrighted content or do
 * anything illegal. This software was designed for academic research
 * purposes and not for illegal activities. We are not responsible for
 * anything illegal you do and by downloading this software you agree
 * not to break any laws. Enjoy!
 *
 * Brian F. Cooper
 * November 21, 2005
 */


#include <stdlib.h>
#include <string.h>
#include "CommunicationManager.h"

#define PLOAD_SIZE 900

const int GatherCommunicationManagerStatistics=1;


/**
 * Statistics about a message type.
 */
class MessageStats
{
  public:
   MessageStats() {count=0; size=0;}

   int count;
   long long size;
};


/**
 * Constructor.
 */
CommunicationManager::CommunicationManager(Runtime * _runtime, ConnectionFactory * _connectionfactory)
{
   runtime=_runtime;
   connectionfactory=_connectionfactory;
   topologymanager=NULL;
   peerlogic=NULL;
   serverconnection=connectionfactory->createServerConnection(runtime->getPeerID(),this);
   connections=new Hashtable(&stringHasher,&stringComparator);
   allconnections=new List();
   curconnection=0;
   //activatedconnections=new Queue();
   messagessent=0;
   bytessent=0;
   messagestats=new Hashtable(&intHasher,&intComparator);

   char buf[1024];
   sprintf(buf,"VIS\t %10s\t %s\t activate", writeTime(runtime->getSystemTime()), runtime->getPeerID()->toString());
   runtime->getLogger()->message(LoggerVis, buf);

   sem_init(&lock,0,1);
}

CommunicationManager::~CommunicationManager()
{
   //terminate();

   Enumeration * e=connections->keys();
   while (e->hasMoreElements())
   {
      char * key=(char*)e->nextElement();
      List * connlist=(List*)connections->get(key);
      delete connlist;
      delete[] key;
   }
   delete e;

   delete allconnections;
   delete connections;
   //delete serverconnection;

   e=messagestats->elements();
   while (e->hasMoreElements())
   {
      MessageStats * ms=(MessageStats*)e->nextElement();
      delete ms;
   }
   delete e;
   delete messagestats;

      

   //delete activatedconnections;

   sem_destroy(&lock);
}

/**
 * Set a topology manager that will decide whether to accept or reject connections, and which will create its own connections.
 */
void CommunicationManager::setTopologyManager(TopologyManager * _topologymanager)
{
   topologymanager=_topologymanager;
}

/**
 * Set a peer logic that will receive incoming messages.
 */
void CommunicationManager::setPeerLogic(PeerLogic * _peerlogic)
{
   peerlogic=_peerlogic;
}

/**
 * Send a message. The message will be deleted after it is sent, so make sure the message destructor cleans up all the state.
 * If there is not a persistent connection of the specified type to the destination, the message will not be sent.
 */
int CommunicationManager::sendMessage(Message * message, int typ)
{
   //sem_wait(&lock);

   PeerID * peer=message->getToPeer();

   List * connlist=(List*)connections->get(peer->toString());
   if (connlist!=NULL)
   {
      Enumeration * e=connlist->elements();
      while (e->hasMoreElements())
      {
	 Connection * c=(Connection*)e->nextElement();
	 if ( (c->getRemotePeerID()->equals(peer)) && (c->getType()==typ) && (c->isValid()) )
	 {
	    //found a connection
	    delete e;

            int mlen = -1;

	    if (GatherCommunicationManagerStatistics)
	    {
	       char * ser=runtime->getMessageSerializer()->serialize(message);
	       mlen=strlen(ser);
	       delete[] ser;

	       bytessent+=mlen;
	       messagessent++;

	       MessageStats * stats=(MessageStats*)messagestats->get((void*)message->getType());
	       if (stats==NULL)
	       {
		  stats=new MessageStats();
		  messagestats->put((void*)message->getType(),(void*)stats);
	       }
	       
	       stats->count++;
	       stats->size+=mlen;
	    }

	    c->sendMessage(message);
  
            char buf[1024];
            char pload[PLOAD_SIZE]; // a little less than buf's size
            char * visstring;

            pload[0] = '\0';

            
            const int PeerQueryMessageType=200;
            const int PeerQueryResultMessageType=201;
            

            /*
            if (PeerQueryMessageType == message->getType())
            {
                char * search = (char *) message->getPayload();
                
                int len = PLOAD_SIZE - 1;
               
                // avoid printing any newlines in the search query.
                // If we print extra newlines the ONA (ODIN Network Activity) file
                // will be unparsable/corrupted.
                char * nlp;
                if (NULL != (nlp = index(search, '\n')))
                {
                    int nloffset = nlp - search;
                    if (nloffset < len) 
                    {
                        len = nloffset;
                    }
                }
                
                // pads pload with several nulls? ugh!
                strncpy(pload, search, len);
            }
            else if (PeerQueryResultMessageType == message->getType())
            {
                List * l = (List *) message->getPayload();

                sprintf(pload, "%6d search results", l->getCount());
            }
            */

            if ( (visstring = message->getVisString()) != NULL){
                strncpy(pload, visstring, PLOAD_SIZE - 1);
                delete visstring;
            }
            
            sprintf(buf,"VIS\t %10s\t %s\t sendmessage\t %s\t %6d\t %6d\t %6d\t %s", writeTime(runtime->getSystemTime()), runtime->getPeerID()->toString(), peer->toString(), typ, message->getType(), mlen, pload);
            runtime->getLogger()->message(LoggerVis, buf);

	    //sem_post(&lock);
	    return 1;
	 }
      }
      delete e;
   }

   //no appropriate connection found
   delete message;

   //sem_post(&lock);
      
   return 0;
}

/**
 * Get a count of the number of messages sent of a particular type.
 */
int CommunicationManager::getMessagesSent(int messagetype)
{
   MessageStats * stats=(MessageStats*)messagestats->get((void*)messagetype);
   if (stats==NULL) return 0;
   
   return stats->count;
}

/**
 * Get a count of the bytes sent for a particular message type.
 */
long long CommunicationManager::getBytesSent(int messagetype)
{
   MessageStats * stats=(MessageStats*)messagestats->get((void*)messagetype);
   if (stats==NULL) return 0;
   
   return stats->size;
}


/**
 * Try to make a persistent connection to a remote peer.
 * The typ parameter indicates the type of overlay link (index link, search link, etc.)
 * This function is asynchronous: the request is made but may be accepted later. The registered topology manager
 * will be notified when the connection is accepted or rejected.
 * If we already have the specified type of connection to that peer, this call will not make a new one, but instead
 * will fail and return 0.
 */
int CommunicationManager::makeConnection(PeerID * peer, int typ)
{
   //sem_wait(&lock);

   List * connlist=(List*)connections->get(peer->toString());
   if (connlist==NULL)
   {
      //no existing connections to that peer
      //make a new list
      connlist=new List();
      char * key=new char[strlen(peer->toString())+1];
      strcpy(key,peer->toString());
      connections->put((void*)key,(void*)connlist);
   }
   else
   {
      Enumeration * e=connlist->elements();
      while (e->hasMoreElements())
      {
	 Connection * c=(Connection*)e->nextElement();
	 if ( (c->getRemotePeerID()->equals(peer)) && (c->getType()==typ) && (c->isValid()) )
	 {
	    //we already have a connection to the specified peer of the given type
	    delete e;
	    
	    //sem_post(&lock);

	    return 0;
	 }
      }
      delete e;
   }
   
   //no existing connection of the specified type to that peer

   //try to make one
   Connection * conn=connectionfactory->requestConnection(runtime->getPeerID(),peer,typ);
   if (conn==NULL)
   {
      //sem_post(&lock);
      return 0;
   }
   
   //success! keep the connection
   connlist->add(conn);
   allconnections->add(conn);

   conn->registerStatusListener(this);

   //the topology manager will want to know when it is closed
   conn->registerConnectionClosedListener(topologymanager);
   //conn->registerConnectionClosedListener(this);

   //sem_post(&lock);
   
   char buf[1024];
   sprintf(buf,"VIS\t %10s\t %s\t connect\t %s\t %6d", writeTime(runtime->getSystemTime()), runtime->getPeerID()->toString(), peer->toString(), typ);
   runtime->getLogger()->message(LoggerVis, buf);

   return 1;
}


/**
 * Close a persistent connection.
 */
void CommunicationManager::closeConnection(PeerID * peer, int typ)
{
   //sem_wait(&lock);
   List * connlist=(List*)connections->get(peer->toString());
   if (connlist==NULL)
   {
      //no existing connections to that peer
      //sem_post(&lock);
      return;
   }

   Enumeration * e=connlist->elements();
   while (e->hasMoreElements())
   {
      Connection * c=(Connection*)e->nextElement();
      if ( (c->getRemotePeerID()->equals(peer)) && (c->getType()==typ) )
      {
	 connlist->remove(c);
	 allconnections->remove(c);

	 delete e;
	 //sem_post(&lock);

	 //connection found. close it
	 if (c->isValid())
	 {
	    c->close();
	 }
	 //the network layer closes it now.
	 //delete c;
	 
	 
	 return;
      }
   }
   delete e;

   //no connection found

   //sem_post(&lock);
}

/**
 * Do one thing; e.g. accept an incoming connection or dispatch a message. If there is something to do, return 1; otherwise, return 0.
 */
/*
int CommunicationManager::executeItem()
{
   //get a connection with something to do
   Connection * conn=(Connection*)activatedconnections->dequeue();
   
   if (conn==NULL) 
   {
      return 0;
   }

   //dispatch messages

   Message * msg=conn->receiveMessage();
   
   if (msg!=NULL)
   {      
      peerlogic->handleMessage(msg,conn->getType());
      return 1;
   }
   
   //if it is ready to close, close it
   if (conn->readyToClose())
   {
      closeConnection(conn->getRemotePeerID(),conn->getType());
      return 1;
   }

   return 0;
}
*/

/*
int CommunicationManager::executeItem()
{
   //sem_wait(&lock);

   //no connections; nothing to do
   if (allconnections->isEmpty())
   {
      //sem_post(&lock);

      return 0;
   }
   
   //make sure we are at a valid index
   if (curconnection>=allconnections->getCount())
   {
      curconnection=0;
   }
   
   int startconnection=curconnection;
   
   //loop until we find a message to dispatch; or, we have tried all connections and found nothing to dispatch
   for (;;)
   {
      //get a connection and look for a message
      Connection * conn=(Connection*)allconnections->getItem(curconnection);

      Message * msg=conn->receiveMessage();
      
      if (msg==NULL)
      {
	 //no message on this connection

	 //look at the next connection

	 curconnection++;
	 if (curconnection>=allconnections->getCount())
	 {
	    curconnection=0;
	 }
	 if (curconnection==startconnection)
	 {
	    //no messages at all
	    //sem_post(&lock);

	    break;
	 }
      }
      else
      {
	 //found a message. dispatch it
	 peerlogic->handleMessage(msg,conn->getType());
	 //sem_post(&lock);
	 return 1;
      }
   }

   //clean up any connections that are ready to close
   List * toclose=new List();
   for (int i=0; i<allconnections->getCount(); i++)
   {
      Connection * conn=(Connection*)allconnections->getItem(i);
      if (conn->readyToClose())
      {
	 toclose->add(conn);
      }
   }
   
   //sem_post(&lock);

   for (int i=0; i<toclose->getCount(); i++)
   {
      Connection * conn=(Connection*)toclose->getItem(i);
      closeConnection(conn->getRemotePeerID(),conn->getType());
   }
   delete toclose;

   return 0;
}
*/

/**
 * Are there things to do? E.g., are there messages queued?
 */
/*
int CommunicationManager::itemsToExecute()
{
  
   return (activatedconnections->size()>0);
}
*/

int CommunicationManager::acceptConnection(PeerID * peer, int typ, Connection * conn)
{
   char buf[4096];

   if (topologymanager==NULL)
   {
      sprintf(buf,"CommunicationManager: cannot accept connection because topology manager is NULL");
      runtime->getLogger()->message(LoggerDebug,buf);
      return 0;
   }

   //sem_wait(&lock);
      
   List * connlist=(List*)connections->get(peer->toString());
   if (connlist!=NULL)
   {
      Enumeration * e=connlist->elements();
      while (e->hasMoreElements())
      {
	 Connection * c=(Connection*)e->nextElement();
	 if ( (c->getRemotePeerID()->equals(peer)) && (c->getType()==typ) && (c->isValid()) )
	 {
	    //we already have a connection to the specified peer of the given type
	    delete e;

	    //sem_post(&lock);

            sprintf(buf,"Rejecting connection because I already have a connection to the specified peer of the given type");
            runtime->getLogger()->message(LoggerDebug,buf);

	    return 0;
	 }
      }
      delete e;
   }
   
   if (topologymanager->acceptConnection(peer,typ))
   {
      //topoology manager accepts connection

      //keep the connection
      if (connlist==NULL)
      {
	 connlist=new List();
	 char * key=new char[strlen(peer->toString())+1];
	 strcpy(key,peer->toString());
	 connections->put(key,connlist);
      }
      connlist->add(conn);
      allconnections->add(conn);
      
      conn->registerStatusListener(this);
	 
      //set the listener for when the connection is closed
      conn->registerConnectionClosedListener(topologymanager);
      conn->registerConnectionClosedListener(this);

      //sem_post(&lock);
      return 1;
   }
   
   //sem_post(&lock);

   sprintf(buf,"Rejecting connection because topology manager told me to");
   runtime->getLogger()->message(LoggerDebug,buf);

   return 0;
}


/**
 * Shut down and clean up.
 */
void CommunicationManager::terminate()
{
   //sem_wait(&lock);

   for (int i=0; i<allconnections->getCount(); i++)
   {
      Connection * c=(Connection*)allconnections->getItem(i);
      c->close();
      
      //the network layer closes it now
      //delete c;
   }

   delete allconnections;
   allconnections=new List();
   
   Enumeration * e=connections->keys();
   while (e->hasMoreElements())
   {
      char * key=(char*)e->nextElement();
      List * connlist=(List*)connections->get(key);
      delete connlist;
      delete[] key;
   }
   delete e;
   delete connections;
   
   connections=new Hashtable(&stringHasher,&stringComparator);

   serverconnection->close();

   //sem_post(&lock);
}



/**
 * Get a list of all of the connections we currently have.
 */
/*
Enumeration * CommunicationManager::getAllConnections()
{
   return allconnections->elements();
}
*/

/**
 * Get a list of all of the neighbors we currently have, as a List * of NeighborInfo objects. Delete the List and NeighborInfo objects when you are done.
 */
List * CommunicationManager::getAllNeighbors()
{
   //sem_wait(&lock);

   //char sbuf[4096];
   //sprintf(sbuf,"* Getting all neighbors (%d connections)",allconnections->getCount());
   //runtime->getLogger()->message(LoggerDebug,sbuf);

   List * ret=new List();
   Enumeration * e=allconnections->elements();
   while (e->hasMoreElements())
   {
      Connection * c=(Connection*)e->nextElement();

      //sprintf(sbuf,"*    Got connection");
      //runtime->getLogger()->message(LoggerDebug,sbuf);

      //sprintf(sbuf,"*    Connection is to %s, type %d",c->getRemotePeerID()->toString(),c->getType());
      //runtime->getLogger()->message(LoggerDebug,sbuf);

      NeighborInfo * ni=new NeighborInfo(c->getRemotePeerID(),c->getType());
      ret->add(ni);

      //sprintf(sbuf,"*    Added connection to list");
      //runtime->getLogger()->message(LoggerDebug,sbuf);
   }
   delete e;

   //sprintf(sbuf,"* Done getting all neighbors");
   //runtime->getLogger()->message(LoggerDebug,sbuf);
   
   //sem_post(&lock);

   return ret;
}

/**
 * Are we currently connected to the specified peer with the specified type?
 */
int CommunicationManager::connectionExists(PeerID * peer, int typ)
{
   //sem_wait(&lock);
   
   List * connlist=(List*)connections->get(peer->toString());
   if (connlist!=NULL)
   {
      Enumeration * e=connlist->elements();
      while (e->hasMoreElements())
      {
	 Connection * c=(Connection*)e->nextElement();
	 if ( (c->getRemotePeerID()->equals(peer)) && (c->getType()==typ) && (c->isValid()) )
	 {
	    //found a connection
	    delete e;
	    //sem_post(&lock);
	    return 1;
	 }
      }
      delete e;
   }

   //no appropriate connection found
   //sem_post(&lock);
      
   return 0;
}


void CommunicationManager::connectionStatusChanged(Connection * conn)
{
   //activatedconnections->enqueue(conn);


   for (;;)
   {
      Message * msg=conn->receiveMessage();
      
      if (msg==NULL)
      {      
	 break;
      }
      
      peerlogic->handleMessage(msg,conn->getType());
   }   

   //if it is ready to close, close it
   if (conn->readyToClose())
   {
      closeConnection(conn->getRemotePeerID(),conn->getType());
   }
}

/**
 * Callback that is executed when a connection closes.
 */
/*
void CommunicationManager::connectionClosed(PeerID * peer, int typ)
{
   //sem_wait(&lock);

   List * connlist=(List*)connections->get(peer->toString());
   if (connlist==NULL)
   {
      //no existing connections to that peer
      //sem_post(&lock);
      return;
   }
   
   Enumeration * e=connlist->elements();
   while (e->hasMoreElements())
   {
      Connection * c=(Connection*)e->nextElement();
      if ( (c->getRemotePeerID()->equals(peer)) && (c->getType()==typ) )
      {
	 //connection found. drop it.
	 connlist->remove(c);
	 allconnections->remove(c);
	 delete c;
	 delete e;
	 //sem_post(&lock);
	 return;
      }
   }
   delete e;
   
   //sem_post(&lock);
}
*/

void CommunicationManager::connectionClosed(PeerID * peer, int typ)
{
   char buf[1024];
   sprintf(buf,"VIS\t %10s\t %s\t disconnect\t %s\t %6d", writeTime(runtime->getSystemTime()), runtime->getPeerID()->toString(), peer->toString(), typ);
   runtime->getLogger()->message(LoggerVis, buf);
}

/**
 * Get a count of the number of messages sent.
 */
int CommunicationManager::getTotalMessagesSent()
{
   return messagessent;
}

/**
 * Get a count of the bytes sent.
 */
long long CommunicationManager::getTotalBytesSent()
{
   return bytessent;
}


/**
 * Get the size, in bytes, used by this CommunicationManager.
 */
int CommunicationManager::getMemorySize()
{
   int size=sizeof(runtime)+sizeof(connectionfactory)+sizeof(peerlogic)+sizeof(topologymanager)+sizeof(serverconnection)+serverconnection->getMemorySize()+
      sizeof(connections)+connections->getMemorySize()+sizeof(allconnections)+allconnections->getMemorySize()+sizeof(curconnection)+sizeof(messagessent)+
      sizeof(bytessent)+sizeof(messagestats)+messagestats->getMemorySize()+sizeof(lock);
   
   Enumeration * e=connections->keys();
   while (e->hasMoreElements())
   {
      char * key=(char*)e->nextElement();
      List * connlist=(List*)connections->get(key);
      size+=strlen(key)+1+connlist->getMemorySize();
   }
   delete e;
   
   for (int i=0; i<allconnections->getCount(); i++)
   {
      Connection * c=(Connection*)allconnections->getItem(i);
      size+=c->getMemorySize();
   }

   e=messagestats->elements();
   while (e->hasMoreElements())
   {
      MessageStats * ms=(MessageStats*)e->nextElement();
      size+=sizeof(*ms);
   }
   delete e;

   return size;
}

   
///////////////////////////////////// NeighborInfo //////////////////////////////////////////

NeighborInfo::NeighborInfo(PeerID * _neighbor, int _typ)
{
   neighbor=_neighbor->clone();
   typ=_typ;
}

NeighborInfo::~NeighborInfo()
{
   delete neighbor;
}

/**
 * Get the connection type.
 */
int NeighborInfo::getType()
{
   return typ;
}

/**
 * Get the peer id of the neighbor.
 */
PeerID * NeighborInfo::getPeerID()
{
   return neighbor;
}
   
