/**
 * ODIN overlay management middleware toolkit
 * 
 * ODIN Group, College of Computing, Georgia Institute of Technology
 *
 * For more information:
 * http://www.cc.gatech.edu/~cooperb/odin/
 * odin.gatech.edu
 *
 * You are free to use, modify and redistribute this software. 
 * However, we make no warranty for this software, and by downloading it
 * you agree not to hold use liable for anything. Do not under any
 * circumstances use this software to share copyrighted content or do
 * anything illegal. This software was designed for academic research
 * purposes and not for illegal activities. We are not responsible for
 * anything illegal you do and by downloading this software you agree
 * not to break any laws. Enjoy!
 *
 * Brian F. Cooper
 * November 21, 2005
 */

#ifndef COMMUNICATIONMANAGER_H
#define COMMUNICATIONMANAGER_H

#include <semaphore.h>
#include "TopologyManager.h"
#include "PeerLogic.h"
#include "PeerID.h"
#include "Connection.h"

class PeerLogic;
class Runtime;
class TopologyManager;

/**
 * Information about a peer's neighbor.
 */
class NeighborInfo
{
  public:
   NeighborInfo(PeerID * _neighbor, int _typ);

   ~NeighborInfo();
   
   /**
    * Get the connection type.
    */
   int getType();
   
   /**
    * Get the peer id of the neighbor.
    */
   PeerID * getPeerID();
   
  private:
   PeerID * neighbor;
   int typ;
};

/**
 * The object that handles sending and receiving messages on the peer side. That is, the PeerLogic sends messages
 * via this object, and this object dispatches received messages to the peer logic.
 */
class CommunicationManager : public ConnectionListener, public ConnectionStatusListener, public ConnectionClosedListener
//, public ConnectionClosedListener
{
  public:
   /************************************ Initialization *************************************************/

   /**
    * Constructor. 
    */
   CommunicationManager(Runtime * _runtime, ConnectionFactory * _connectionfactory);

   virtual ~CommunicationManager();

   /**
    * Set a topology manager that will decide whether to accept or reject connections, and which will create its own connections.
    */
   void setTopologyManager(TopologyManager * _topologymanager);
   
   /**
    * Set a peer logic that will receive incoming messages.
    */
   void setPeerLogic(PeerLogic * _peerlogic);

   /************************************ Requests from the peer layer *************************************/
   
   /**
    * Send a message. The message will be deleted after it is sent, so make sure the message destructor cleans up all the state.
    * If there is not a persistent connection of the specified type the destination, the message will not be sent. Returns 1 if the message sent, 0 if not.
    */
   int sendMessage(Message * message, int typ);

   /**
    * Try to make a persistent connection to a remote peer.
    * The typ parameter indicates the type of overlay link (index link, search link, etc.)
    * Returns 1 if the connection is accepted, 0 if not.
    */
   int makeConnection(PeerID * peer, int typ);

   /**
    * Close a persistent connection.
    */
   void closeConnection(PeerID * peer, int typ);

   /**
    * Get a list of all of the neighbors we currently have, as a List * of NeighborInfo objects. Delete the List and NeighborInfo objects when you are done.
    */
   List * getAllNeighbors();

   /**
    * Are we currently connected to the specified peer with the specified type?
    */
   int connectionExists(PeerID * peer, int typ);

   /**
    * Get a count of the number of messages sent.
    */
   int getTotalMessagesSent();
   
   /**
    * Get a count of the bytes sent.
    */
   long long getTotalBytesSent();

   /**
    * Get a count of the number of messages sent of a particular type.
    */
   int getMessagesSent(int messagetype);
   
   /**
    * Get a count of the bytes sent for a particular message type.
    */
   long long getBytesSent(int messagetype);

   /*********************************** Driving by the runtime ******************************************/

   
   /**
    * Do one thing; e.g. accept an incoming connection or dispatch a message. If there is something to do, return 1; otherwise, return 0.
    */
   //int executeItem();

   /**
    * Are there things to do? E.g., are there messages queued?
    */
   //int itemsToExecute();

   /**
    * Shut down and clean up.
    */
   void terminate();

   /*********************************** from ConnectionListener *****************************************/
   
   /**
    * Should we accept a connection of the specified type from the remote peer. If so, the conn object will be the connection, and return 1. If not, return 0.
    */
   virtual int acceptConnection(PeerID * peer, int typ, Connection * conn);

   /********************************** from ConnectionClosedListener ***********************************/

   /**
    * Callback that is executed when a connection closes.
    */
   virtual void connectionClosed(PeerID * peer, int typ);
   
   /********************************** from ConnectionStatusListener ***********************************/

   /**
    * Callback that is called when a connection has changed status (e.g. a new message has arrived or the connection is ready to close.
    */
   virtual void connectionStatusChanged(Connection * conn);

   /******************************** Diagnostic ***********************************/

   /**
    * Get the size, in bytes, used by this CommunicationManager.
    */
   virtual int getMemorySize();

  private:
   Runtime * runtime;
   ConnectionFactory * connectionfactory;
   PeerLogic * peerlogic;
   TopologyManager * topologymanager;
   ServerConnection * serverconnection;
   Hashtable * connections;
   List * allconnections;
   int curconnection;
   //Queue * activatedconnections;

   //statistics
   int messagessent;
   long long bytessent;
   Hashtable * messagestats;

   sem_t lock;
};

#endif
