/**
 * ODIN overlay management middleware toolkit
 * 
 * ODIN Group, College of Computing, Georgia Institute of Technology
 *
 * For more information:
 * http://www.cc.gatech.edu/~cooperb/odin/
 * odin.gatech.edu
 *
 * You are free to use, modify and redistribute this software. 
 * However, we make no warranty for this software, and by downloading it
 * you agree not to hold use liable for anything. Do not under any
 * circumstances use this software to share copyrighted content or do
 * anything illegal. This software was designed for academic research
 * purposes and not for illegal activities. We are not responsible for
 * anything illegal you do and by downloading this software you agree
 * not to break any laws. Enjoy!
 *
 * Brian F. Cooper
 * November 21, 2005
 */

#ifndef MESSAGE_H
#define MESSAGE_H

#include "PeerID.h"
#include "utils/List.h"

const int SimpleMessageType=0;

/**
 * A message that is sent over the network between peers. This object is a high
 * level representation of a message. Each message has an application-specified type. Users of the ODINSim package
 * should take care to ensure that message types are unique. Message itself is abstract. ODINSim
 * provides a single basic message, SimpleMessage, that has a char * string as a payload. This
 * message type is SimpleMessageType=0. Therefore, any subclasses of Message that you define
 * should have a type other than 0. 
 *
 * The MessageSerializer can take a message and 
 * serialize it to a wire format; and conversely can take a wireformat message and 
 * unserialize it. However, this means that any subclasses of message should also 
 * register a serializer with the global MessageSerializer to provide special handling 
 * of the payload. 
 *
 * The type of different messages is representing using an integer type, rather than by asking
 * for the objects actual class (e.g. using dynamic_cast) so that you can use Message directly if you like;
 * that is, you can create different types of messages without necessarily subclassing message. This is 
 * useful for simple messages that aren't necessarily strings (e.g. SimpleMessage).
 */
class Message
{
  public:
   
   /**
    * Create a message with a single destination. Clones are made of the PeerIDs so you can delete those after creating this message if you like. 
    * The payload is not cloned and won't be deleted when the message is deleted. Subclasses, which know the type of _payload, may have different behavior. 
    */
   Message(PeerID * _to, PeerID * _from, int _typ, void * _payload);
   
   virtual ~Message();
   
   /**
    * Get the id of the peer this message is addressed to.
    */
   virtual PeerID * getToPeer();

   /**
    * Get the id of the peer that sent this message.
    */
   virtual PeerID * getFromPeer();

   /**
    * Get the type of this message.
    */
   virtual int getType();

   /**
    * Get the payload of this message.
    */
   virtual void * getPayload();

   /**
    * Get the visualization string for this message.
    */
   virtual char * getVisString();

  private:
   PeerID * to;
   PeerID * from;
   int typ;
   void * payload;
};

/**
 * A simple subclass of Message. The payload is a char * string.
 */
class SimpleMessage : public Message
{
  public:
   /**
    * Create a message with a single destination. The _payload parameter is copied, so you can delete it after creating the message. 
    * The peerids are also cloned so you can delete them if you like.
    */
   SimpleMessage(PeerID * _to, PeerID * _from, char * _payload);

   virtual ~SimpleMessage();
  
   virtual char * getStringPayload();
};



#endif
