/**
 * ODIN overlay management middleware toolkit
 * 
 * ODIN Group, College of Computing, Georgia Institute of Technology
 *
 * For more information:
 * http://www.cc.gatech.edu/~cooperb/odin/
 * odin.gatech.edu
 *
 * You are free to use, modify and redistribute this software. 
 * However, we make no warranty for this software, and by downloading it
 * you agree not to hold use liable for anything. Do not under any
 * circumstances use this software to share copyrighted content or do
 * anything illegal. This software was designed for academic research
 * purposes and not for illegal activities. We are not responsible for
 * anything illegal you do and by downloading this software you agree
 * not to break any laws. Enjoy!
 *
 * Brian F. Cooper
 * November 21, 2005
 */


#include <stdlib.h>
#include "Runtime.h"

Runtime::Runtime(PeerID * _peerid, Logger * _logger, Measurements * _measurements)
{
   peerid=_peerid->clone();
   logger=_logger;
   measurements=_measurements;
   listeners=new List();
   topologymanager=NULL;
   peerlogic=NULL;
}


Runtime::~Runtime()
{
   delete peerid;
   delete listeners;   
   //delete timerqueue;

   /*
   while (eventqueue->size())
   {
      Event * ev=(Event*)eventqueue->dequeue();
      delete ev;
   }
   */
   
   //delete eventqueue;

}

/**
 * Get the PeerID for this peer.
 */
PeerID * Runtime::getPeerID()
{
   return peerid;
}


/**
 * Get the information message logger for the system.
 */
Logger * Runtime::getLogger()
{
   return logger;
}

/**
 * Assign a TopologyManager instance for this peer.
 */
void Runtime::setTopologyManager(TopologyManager * _topologymanager)
{
   topologymanager=_topologymanager;
   getCommunicationManager()->setTopologyManager(topologymanager);
}


/**
 * Get the object to report measurements to.
 */
Measurements * Runtime::getMeasurements()
{
   return measurements;
}

/**
 * Assign a PeerLogic instance for this peer.
 */
void Runtime::setPeerLogic(PeerLogic * _peerlogic)
{
   peerlogic=_peerlogic;
   getCommunicationManager()->setPeerLogic(peerlogic);
}


/**
 * Handle a timer event.
 */
void Runtime::handleTimer(void * data)
{
   peerlogic->handleTimer(data);
}

/**
 * Register an event listener. Event listeners receive events from the peer. Applications should register themselves with the runtime using this method.
 */
void Runtime::registerEventListener(EventListener * listener)
{
   listeners->add((void*)listener);
}

/**
 * Send an event to the peer. The application should call this function to enqueue an event for the PeerLogic to handle.
 * The event will be cloned, so you can delete it after you are done calling this method.
 */
void Runtime::sendEventToPeer(Event * ev)
{
   //eventqueue->enqueue((void*)ev->clone());
   peerlogic->handleEvent(ev->clone());
}


/**
 * Send an event from the peer to the application. Any registered listeners will receive a copy of the event. The event will be cloned, so you can delete the original.
 */
void Runtime::sendEventToApplication(Event * ev)
{
   Enumeration * e=listeners->elements();
   
   while (e->hasMoreElements())
   {
      EventListener * l=(EventListener*)e->nextElement();
      l->handleEvent(ev->clone());
   }
   delete e;
}


/**
 * Terminate this peer and cleanup. Peers should call shutdown() to terminate themselves. This function should only be called by the PeerExecutive.
 */
void Runtime::terminate()
{
   //XXX
   //fprintf(stderr,"Peer %s: %d bytes\n",getPeerID()->toString(),getMemorySize());
   
   getPeerLogic()->terminate();
   getTopologyManager()->terminate();
   getCommunicationManager()->terminate();

   char buf[1024];
   sprintf(buf,"VIS\t %10s\t %s\t deactivate", writeTime(getSystemTime()), getPeerID()->toString());
   getLogger()->message(LoggerVis, buf);
}

/**
 * Get the PeerLogic for this peer.
 */
PeerLogic * Runtime::getPeerLogic()
{
   return peerlogic;
}

/**
 * Get the topology manager for this peer.
 */
TopologyManager * Runtime::getTopologyManager()
{
   return topologymanager;
}

/**
 * Get the size, in bytes, used by this peer.
 */
int Runtime::getMemorySize()
{
   int size=sizeof(peerid)+sizeof(*peerid)+sizeof(logger)+sizeof(measurements)+sizeof(peerlogic)+sizeof(listeners)+sizeof(topologymanager);
   
   size+=listeners->getMemorySize();
   size+=peerlogic->getMemorySize();
   //size+=topologymanager->getMemorySize();

   return size;
}
