/**
 * ODIN overlay management middleware toolkit
 * 
 * ODIN Group, College of Computing, Georgia Institute of Technology
 *
 * For more information:
 * http://www.cc.gatech.edu/~cooperb/odin/
 * odin.gatech.edu
 *
 * You are free to use, modify and redistribute this software. 
 * However, we make no warranty for this software, and by downloading it
 * you agree not to hold use liable for anything. Do not under any
 * circumstances use this software to share copyrighted content or do
 * anything illegal. This software was designed for academic research
 * purposes and not for illegal activities. We are not responsible for
 * anything illegal you do and by downloading this software you agree
 * not to break any laws. Enjoy!
 *
 * Brian F. Cooper
 * November 21, 2005
 */


#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <netdb.h>
#include "platform/PlatformRuntime.h"
#include "platform/SocketManager.h"
#include "platform/PlatformController.h"
#include "peer/Runtime.h"
#include "peer/PeerLogic.h"
#include "peer/Logger.h"
#include "peer/TopologyManager.h"
#include "peer/IPPeerID.h"
#include "auxil/MessageSerializer.h"
#include "utils/Hashtable.h"
#include "utils/StringBuffer.h"
#include "common/Hostcatcher.h"
#include "kernel/TraceWorkloadExecutive.h"
#include "rwalksim/RandomWalkPeerLogicFactory.h"
#include "rwalksim/MinDegreeTopologyManager.h"
#include "rwalksim/SquareRootTopologyManager.h"
#include "rwalksim/MinSearchIndexDegreeTopologyManager.h"
#include "rwalksim/BasicPeerLogic.h"
#include "common/ExtensiblePeerLogicFactory.h"
#include "common/ExtensibleTopologyManagerFactory.h"
#include "WordIndexQueryProcessor.h"

const char * RequiredProperties[] = 
{
   "hostcatcheraddress",
   "peerlogictype",
   "topologymanagertype",
};

const char * PropertyDescriptions[] =
{
   "The PeerID and port of the hostcatcher",
   "The kind of peer logic to create",
   "The kind of topology manager to create",
};

int RequiredPropertiesCount=3;

void usageMessage(char * arg0)
{
   printf("Usage: %s [-v n] [-r n] [-m] [-i indexfile] [-p property=value] property-file peer-port control-port\n",arg0);
   printf("Options: \n");
   printf("  -comment \"comment\": make a comment about this run\n");
   printf("  -v n: use logging level n (0 is silent, 4 is most verbose)\n");
   printf("  -r n: use random seed n\n");
   printf("  -m:   gather measurements\n");
   printf("  -i:   process queries using the specified indexfile\n");
   printf("  -p:   specify a property value that will add to or override any in the property file\n");
}

int main(int argc, char ** argv)
{
   unsigned int seed=time(NULL);
   
   int argindex=1;
   int debuglevel=LoggerUrgent;
   int gathermeasurements=0;
   char * indexfile=NULL;

   Properties * prop=new Properties();

   //parse command line options
   while ( (argindex<argc) && (argv[argindex][0]=='-') )
   {
      if (strcmp(argv[argindex],"-v")==0)
      {
	 argindex++;
	 debuglevel=atoi(argv[argindex]);
	 argindex++;
      }
      else if (strcmp(argv[argindex],"-r")==0)
      {
	 argindex++;
	 seed=atoi(argv[argindex]);
	 argindex++;
      }
      else if (strcmp(argv[argindex],"-m")==0)
      {
	 argindex++;
	 gathermeasurements=1;
      }
      else if (strcmp(argv[argindex],"-i")==0)
      {
	 argindex++;
	 indexfile=argv[argindex];
	 argindex++;
      }
      else if (strcmp(argv[argindex],"-p")==0)
      {
	 argindex++;
	 char * val=strstr(argv[argindex],"=");
	 if (val!=NULL)
	 {
	    *val=0;
	    val++;
	    prop->setProperty(argv[argindex],val);
	    val--;
	    *val='=';
	 }
	 argindex++;
      }
      else
      {
	 usageMessage(argv[0]);
	 return 0;
      }
   }

   if (argindex+3!=argc)
   {
      usageMessage(argv[0]);
      return 0;
   }

   srandom(seed);

   char * propertyfile=argv[argindex];
   argindex++;

   int port=atoi(argv[argindex]);
   argindex++;

   int controlport=atoi(argv[argindex]);
   argindex++;
   
   // Load properties
   Properties * properties=new Properties(propertyfile);

   //add any command line properties
   Enumeration * e=prop->getProperties();
   while (e->hasMoreElements())
   {
      char * k=(char*)e->nextElement();
      char * v=prop->getProperty(k);
      properties->setProperty(k,v);
      //printf("Setting property [%s]=[%s]\n",k,v);
   }
   delete e;

   //check for required properties
   for (int i=0; i<RequiredPropertiesCount; i++)
   {
      if (properties->getProperty(RequiredProperties[i])==NULL)
      {
	 printf("Missing property: %s (%s)\n",RequiredProperties[i],PropertyDescriptions[i]);
	 return 0;
      }
   }

   // Set up platform
   
   //first, get local host's peer id
   char hnamebuf[2048];
   gethostname(hnamebuf,2048);
   struct hostent * haddr=gethostbyname(hnamebuf);
   IPPeerID * pid=new IPPeerID(haddr->h_addr[0],haddr->h_addr[1],haddr->h_addr[2],haddr->h_addr[3],port);
   
   printf("My peerid: %s\n",pid->toString());


   //set up runtime
   Logger * logger=new Logger(debuglevel);
   Measurements * meas=new Measurements(argc,argv,0,gathermeasurements);

   MessageSerializer * mserializer=new MessageSerializer();
   mserializer->registerSingleSerializer(new HostcatcherMessageSerializer(mserializer));
   mserializer->registerSingleSerializer(new PeerMessageSerializer(mserializer));
   mserializer->registerObjectSerializer(new WordIndexContentIndexSerializer(mserializer));

   //set up query processor
   /*
   ContentMap * contentmap=NULL;

   if (properties->getProperty("contentmapfile")!=NULL)
   {
      contentmap=new ContentMap(properties->getProperty("contentmapfile"),NULL);
      mserializer->registerObjectSerializer(new ContentMapContentIndexSerializer(contentmap,mserializer));
   }
   */
   WordIndexQueryProcessorFactory * wiqpf=NULL;
   if (indexfile!=NULL)
   {
      wiqpf=new WordIndexQueryProcessorFactory(indexfile,0.1,logger);
   }

   SocketManager * socketmanager=new SocketManager();

   PlatformRuntime * runtime=new PlatformRuntime(pid,socketmanager,logger,meas,mserializer);

   socketmanager->setRuntime(runtime);

   PlatformController * controller=new PlatformController(runtime,socketmanager,controlport);

   //set up factories
   ExtensiblePeerLogicFactory * eplf=new ExtensiblePeerLogicFactory();
   eplf->registerFactory(new HostcatcherPeerLogicFactory());
   if (wiqpf!=NULL)
   {
      eplf->registerFactory(new BasicPeerLogicFactory(wiqpf));
      eplf->registerFactory(new RandomWalkPeerLogicFactory(wiqpf,logger));
   }

   ExtensibleTopologyManagerFactory * etmf=new ExtensibleTopologyManagerFactory();
   etmf->registerFactory(new HostcatcherTopologyManagerFactory());
   etmf->registerFactory(new MinDegreeTopologyManagerFactory(logger));
   etmf->registerFactory(new SquareRootTopologyManagerFactory(logger));
   etmf->registerFactory(new MinSearchIndexDegreeTopologyManagerFactory(logger));

   //create peer logic and topology manager
   PeerLogic * pl=eplf->newPeerLogic(runtime,properties);
   runtime->setPeerLogic(pl);
   
   TopologyManager * tm=etmf->newTopologyManager(runtime,properties);
   runtime->setTopologyManager(tm);
   
   Schema * messagestatistics=new Schema("MessageStatistics");
   messagestatistics->addField("Peer",TypeString);
   messagestatistics->addField("MessageType",TypeInt);
   messagestatistics->addField("MessagesSent",TypeInt);
   messagestatistics->addField("BytesSent",TypeLongLong);
   meas->newSchema(messagestatistics);

   Schema * degreestatistics=new Schema("DegreeStatistics");
   degreestatistics->addField("Peer",TypeString);
   degreestatistics->addField("SearchConnections",TypeInt);
   degreestatistics->addField("IndexConnections",TypeInt);
   meas->newSchema(degreestatistics);

   Schema * squarerootstatistics=new Schema("SquareRootStatistics");
   squarerootstatistics->addField("Peer",TypeString);
   squarerootstatistics->addField("TotalQueries",TypeInt);
   squarerootstatistics->addField("MatchedQueries",TypeInt);
   squarerootstatistics->addField("IdealDegree",TypeInt);
   meas->newSchema(squarerootstatistics);

   if (gathermeasurements)
   {
      meas->printRun();
      meas->printSchema("MessageStatistics");
      meas->printSchema("DegreeStatistics");
      meas->printSchema("SquareRootStatistics");
   }

   pl->init();
   tm->init();
   controller->init();
   
   runtime->run();
  
   
 
   //clean up
   delete messagestatistics;
   delete degreestatistics;
   delete squarerootstatistics;

   eplf->deleteAllFactories();
   delete eplf;
   etmf->deleteAllFactories();
   delete etmf;

   delete runtime;
   delete pl;
   delete tm;

   delete meas;
   delete wiqpf;
   delete properties;
   delete logger;
   mserializer->deleteAllSerializers();
   delete mserializer;


   return 0;
}
