/**
 * ODIN overlay management middleware toolkit
 * 
 * ODIN Group, College of Computing, Georgia Institute of Technology
 *
 * For more information:
 * http://www.cc.gatech.edu/~cooperb/odin/
 * odin.gatech.edu
 *
 * You are free to use, modify and redistribute this software. 
 * However, we make no warranty for this software, and by downloading it
 * you agree not to hold use liable for anything. Do not under any
 * circumstances use this software to share copyrighted content or do
 * anything illegal. This software was designed for academic research
 * purposes and not for illegal activities. We are not responsible for
 * anything illegal you do and by downloading this software you agree
 * not to break any laws. Enjoy!
 *
 * Brian F. Cooper
 * November 21, 2005
 */


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "PeerMessages.h"
#include "utils/StringBuffer.h"

#define VIS_SIZE 900
////////////////////////////////////// PeerQueryMessage //////////////////////////////////////////////////

/**
 * Create a search message. Each search has a unique id (_messageid) that is retained as the message is forwarded.
 * _searchingpeer and _query are cloned so you can delete them.
 */
PeerQueryMessage::PeerQueryMessage(PeerID * _to, PeerID * _from, char * _query, PeerID * _searchingpeer, int _messageid, int _ttl, int _resultcount) :
   Message(_to,_from,PeerQueryMessageType,NULL)
{
   query=new char[strlen(_query)+1];
   strcpy(query,_query);
   searchingpeer=_searchingpeer->clone();
   messageid=_messageid;
   ttl=_ttl;
   resultcount=_resultcount;
}

PeerQueryMessage::~PeerQueryMessage()
{
   delete[] query;
   delete searchingpeer;
}
   
/**
 * Get the query. Don't delete this string.
 */
char * PeerQueryMessage::getQuery()
{
   return query;
}
   
/**
 * Get the id of the original searching peer.
 */
PeerID * PeerQueryMessage::getSearchingPeerID()
{
   return searchingpeer;
}
   
/**
 * Get the id of this particular search.
 */
int PeerQueryMessage::getMessageID()
{
   return messageid;
}
   
/**
 * Get the time-to-live of this message.
 */
int PeerQueryMessage::getTTL()
{
   return ttl;
}
   
/**
 * Get the target result count, or -1 if none.
 */
int PeerQueryMessage::getResultCount()
{
   return resultcount;
}

/**
 * Get the visualization string for this message.
 */
char * PeerQueryMessage::getVisString()
{
    char * search = query;
    char * visstring = new char[VIS_SIZE];
    visstring[0] = '\0';

    int len = VIS_SIZE - 1;

    // avoid printing any newlines in the search query.
    // If we print extra newlines the ONA (ODIN Network Activity) file
    // will be unparsable/corrupted.
    char * nlp;
    if (NULL != (nlp = index(search, '\n')))
    {
        int nloffset = nlp - search;
        if (nloffset < len)
        {
            len = nloffset;
        }
    }

    // pads visstring with several nulls? ugh!
    strncpy(visstring, search, len);
    return visstring;
}


    

//////////////////////////////////////// PeerQueryResultMessage //////////////////////////////////////////////


/**
 * Create a result message. The list of results will be kept (e.g. not cloned) and deleted when the message is deleted.
 */
PeerQueryResultMessage::PeerQueryResultMessage(PeerID * _to, PeerID * _from, int _messageid, List * _results) : Message(_to,_from,PeerQueryResultMessageType,NULL)
{
   messageid=_messageid;
   results=_results;
}
   
PeerQueryResultMessage::~PeerQueryResultMessage()
{
   List * l=getResults();
   
   if (l!=NULL)
   {
      
      for (int i=0; i<l->getCount(); i++)
      {
	 QueryResult * qr=(QueryResult*)l->getItem(i);
	 delete qr;
      }
      delete l;
   }
}

   
/**
 * Get the results, and remove a reference to the list from this message object. Then, when the message
 * gets deleted, the list won't be deleted. This allows you to pass the results list to another object
 * without having to copy it. But future calls to extractResults() or getResults() on this object will return NULL.
 */
List * PeerQueryResultMessage::extractResults()
{
   List * ret=results;
   results=NULL;
   
   return ret;
}
   
/**
 * Get the results, as a List of QueryResult objects.
 */
List * PeerQueryResultMessage::getResults()
{
   return results;
}

int PeerQueryResultMessage::getMessageID()
{
   return messageid;
}

char * PeerQueryResultMessage::getVisString()
{
    char * visstring = new char[32];
    List * l = results;

    sprintf(visstring, "%6d search results", l->getCount());
    
    return visstring;
}


////////////////////////////////////// PeerIndexMessage //////////////////////////////////////

PeerIndexMessage::PeerIndexMessage(PeerID * _to, PeerID * _from, ContentIndex * _index) : Message(_to,_from,PeerIndexMessageType,(void*)_index)
{
}

PeerIndexMessage::~PeerIndexMessage()
{
   delete getIndex();
}

ContentIndex * PeerIndexMessage::getIndex()
{
   return (ContentIndex*)getPayload();
}
   

/////////////////////////////////////// PeerIndexRemoveMessage /////////////////////////////////

PeerIndexRemoveMessage::PeerIndexRemoveMessage(PeerID * _to, PeerID * _from) : Message(_to,_from,PeerIndexRemoveMessageType,NULL) 
{
}

PeerIndexRemoveMessage::PeerIndexRemoveMessage(PeerID * _to, PeerID * _from, char * _indexname) : Message(_to,_from,PeerIndexRemoveMessageType,new char[strlen(_indexname)+1]) 
{
   char * name=(char*)getPayload();
   
   strcpy(name,_indexname);
}

/**
 * Get the name of the index to invalidate, or NULL if we want to invalidate all of them.
 */
char * PeerIndexRemoveMessage::getIndexName()
{
   return (char*)getPayload();
}

/////////////////////////////////////// PeerMessageSerializer /////////////////////////////////

/**
 * Create a PeerMessageSerializer. The MessageSerializer given in this constructor should be able to serialize ContentIndex objects.
 */
PeerMessageSerializer::PeerMessageSerializer(MessageSerializer * _serializer)
{
   serializer=_serializer;
}
   
/**
 * Returns 1 if this instance of SingleSerializer knows how to serialize the given message. Otherwise, returns 0.
 */
int PeerMessageSerializer::canSerialize(Message * m)
{
   if ( (dynamic_cast<PeerQueryMessage*>(m)) ||
	(dynamic_cast<PeerQueryResultMessage*>(m)) ||
        (dynamic_cast<PeerIndexMessage*>(m)) || 
	(dynamic_cast<PeerIndexRemoveMessage*>(m)) )
   {
      return 1;
   }
   
   return 0;
}
   
/**
 * Returns 1 if this instance of SingleSerializer knows how to deserialize the given message. Otherwise, returns 0.
 */
int PeerMessageSerializer::canDeserialize(char * buf)
{
   if ( (strncmp(buf,PeerQueryMessageMagic,strlen(PeerQueryMessageMagic))==0) ||
	(strncmp(buf,PeerQueryResultMessageMagic,strlen(PeerQueryResultMessageMagic))==0) ||
	(strncmp(buf,PeerIndexMessageMagic,strlen(PeerIndexMessageMagic))==0) || 
	(strncmp(buf,PeerIndexRemoveMessageMagic,strlen(PeerIndexRemoveMessageMagic))==0) )
   {
      return 1;
   }

   return 0;
}
   
/**
 * Get a buffer containing the serialized version of the message. 
 * Delete the buffer when you are done with it.
 */
char * PeerMessageSerializer::serialize(Message * m)
{
   if (dynamic_cast<PeerQueryMessage*>(m)) 
   {
      StringBuffer * buf=new StringBuffer();
      PeerQueryMessage * pqm=(PeerQueryMessage*)m;
      
      buf->addString(PeerQueryMessageMagic);
      
      char * pid=serializer->serializePeerID(m->getToPeer());
      buf->addString(pid);
      delete[] pid;
      
      pid=serializer->serializePeerID(m->getFromPeer());
      buf->addString(pid);
      delete[] pid;

      pid=serializer->serializePeerID(pqm->getSearchingPeerID());
      buf->addString(pid);
      delete[] pid;

      char tbuf[256];
      sprintf(tbuf,"%d|%d|%d|",pqm->getMessageID(),pqm->getTTL(),pqm->getResultCount());
      buf->addString(tbuf);
      
      buf->addString(pqm->getQuery());
      
      char * ser=buf->getString();
      
      delete buf;
      
      return ser;
   }
   else if (dynamic_cast<PeerQueryResultMessage*>(m)) 
   {
      StringBuffer * buf=new StringBuffer();
      PeerQueryResultMessage * pqrm=(PeerQueryResultMessage*)m;
      
      buf->addString(PeerQueryResultMessageMagic);
      
      char * pid=serializer->serializePeerID(m->getToPeer());
      buf->addString(pid);
      delete[] pid;
      
      pid=serializer->serializePeerID(m->getFromPeer());
      buf->addString(pid);
      delete[] pid;

      buf->addInt(pqrm->getMessageID());
      buf->addString("|");

      List * l=pqrm->getResults();
      
      for (int i=0; i<l->getCount(); i++)
      {
	 QueryResult * qr=(QueryResult*)l->getItem(i);
	 
	 pid=serializer->serializePeerID(qr->getPeer());
	 buf->addString(pid);
	 delete[] pid;
	 
	 buf->addString(qr->getData());
	 buf->addString("|");
      }

      char * ser=buf->getString();
      
      delete buf;
      
      return ser;
   }
   else if (dynamic_cast<PeerIndexMessage*>(m)) 
   {
      PeerIndexMessage * pim=(PeerIndexMessage*)m;

      StringBuffer * buf=new StringBuffer();
      buf->addString(PeerIndexMessageMagic);
      char * pidser=serializer->serializePeerID(m->getToPeer());
      buf->addString(pidser);
      delete[] pidser;
      
      pidser=serializer->serializePeerID(m->getFromPeer());
      buf->addString(pidser);
      delete[] pidser;
      
      char * iser=serializer->serializeObject(pim->getIndex());
      buf->addString(iser);
      delete[] iser;
      
      char * ret=buf->getString();
      delete buf;
      return ret;
   }
   else if (dynamic_cast<PeerIndexRemoveMessage*>(m)) 
   {
      StringBuffer * buf=new StringBuffer();
      buf->addString(PeerIndexRemoveMessageMagic);
      char * pidser=serializer->serializePeerID(m->getToPeer());
      buf->addString(pidser);
      delete[] pidser;
      
      pidser=serializer->serializePeerID(m->getFromPeer());
      buf->addString(pidser);
      delete[] pidser;

      char * ret=buf->getString();
      delete buf;
      return ret;
   }

   return NULL;
}
   
/**
 * Deserialize a message buffer into a message. 
 */
Message * PeerMessageSerializer::deserialize(char * buf)
{
   if (strncmp(buf,PeerQueryMessageMagic,strlen(PeerQueryMessageMagic))==0) 
   {
      char * ptr=buf+=strlen(PeerQueryMessageMagic);
      
      int len;

      PeerID * topid=serializer->deserializePeerID(ptr,len);
      ptr+=len;

      PeerID * frompid=serializer->deserializePeerID(ptr,len);
      ptr+=len;

      PeerID * searchingpid=serializer->deserializePeerID(ptr,len);
      ptr+=len;

      int mid=atoi(ptr);
      ptr=strstr(ptr,"|");
      ptr++;
      
      int ttl=atoi(ptr);
      ptr=strstr(ptr,"|");
      ptr++;
      
      int rcount=atoi(ptr);
      ptr=strstr(ptr,"|");
      ptr++;
      
      Message * ret=new PeerQueryMessage(topid,frompid,ptr,searchingpid,mid,ttl,rcount);
      delete topid;
      delete frompid;
      delete searchingpid;
      
      return ret;
   }
   else if (strncmp(buf,PeerQueryResultMessageMagic,strlen(PeerQueryResultMessageMagic))==0) 
   {
      char * ptr=buf+=strlen(PeerQueryResultMessageMagic);
      
      int len;

      PeerID * topid=serializer->deserializePeerID(ptr,len);
      ptr+=len;

      PeerID * frompid=serializer->deserializePeerID(ptr,len);
      ptr+=len;

      int mid=atoi(ptr);
      ptr=strstr(ptr,"|");
      ptr++;

      List * results=new List();
      
      while (*ptr)
      {
	 PeerID * rpid=serializer->deserializePeerID(ptr,len);
	 ptr+=len;
	 char * tptr=strstr(ptr,"|");
	 *tptr=0;
	 QueryResult * qr=new QueryResult(ptr,rpid);
	 results->add(qr);
	 delete rpid;
	 *tptr='|';
	 ptr=tptr+1;
      }

      Message * ret=new PeerQueryResultMessage(topid,frompid,mid,results);
      
      delete topid;
      delete frompid;
      
      return ret;
   }
   else if (strncmp(buf,PeerIndexMessageMagic,strlen(PeerIndexMessageMagic))==0) 
   {
      char * ptr=buf+=strlen(PeerIndexMessageMagic);
      int len;
      PeerID * topid=serializer->deserializePeerID(ptr,len);
      ptr+=len;
      PeerID * frompid=serializer->deserializePeerID(ptr,len);
      ptr+=len;
      ContentIndex * ci=(ContentIndex*)serializer->deserializeObject(ptr,len);
      PeerIndexMessage * pim=new PeerIndexMessage(topid,frompid,ci);
      delete topid;
      delete frompid;
      
      return pim;
   }
   else if (strncmp(buf,PeerIndexRemoveMessageMagic,strlen(PeerIndexRemoveMessageMagic))==0) 
   {
      char * ptr=buf+strlen(PeerIndexRemoveMessageMagic);
      int len;
      PeerID * topid=serializer->deserializePeerID(ptr,len);
      ptr+=len;
      PeerID * frompid=serializer->deserializePeerID(ptr,len);
      PeerIndexRemoveMessage * pirm=new PeerIndexRemoveMessage(topid,frompid);
      delete topid;
      delete frompid;
      return pirm;
   }

   return NULL;
}   

/*
#include "peer/IPPeerID.h"
#include "ContentMap.h"
#include "ContentMapQueryProcessor.h"

int main(int argc, char ** argv)
{
   ContentMap *cm=new ContentMap(argv[1],NULL);
   MessageSerializer * mser=new MessageSerializer();
   
   mser->registerObjectSerializer(new ContentMapContentIndexSerializer(cm,mser));

   PeerMessageSerializer * pms=new PeerMessageSerializer(mser);
   
   PeerID * pid1=new IPPeerID(10,20,30,40,2000);
   PeerID * pid2=new IPPeerID(50,60,70,80,2000);
   PeerID * pid3=new IPPeerID(90,100,110,120,2000);

   char buf[256];
   strcpy(buf,"APPLE BANANA");
   PeerQueryMessage * pqm=new PeerQueryMessage(pid1,pid2,buf,pid3,42,7,12);
   
   char * ser=pms->serialize(pqm);
   
   printf("[%s]\n",ser);

   PeerQueryMessage * pqm2=(PeerQueryMessage*)pms->deserialize(ser);
   
   printf("Message:\nTo: %s\nFrom: %s\nSearching peer: %s\nMessage id: %d\nTTL: %d\nResults: %d\nQuery: \"%s\"\n",pqm2->getToPeer()->toString(),pqm2->getFromPeer()->toString(),pqm2->getSearchingPeerID()->toString(),pqm2->getMessageID(),pqm2->getTTL(),pqm2->getResultCount(),pqm2->getQuery());
   
   List * res=new List();
   strcpy(buf,"a.dat");
   QueryResult * qr=new QueryResult(buf,pid2);
   res->add(qr);

   strcpy(buf,"bb.dat");
   qr=new QueryResult(buf,pid3);
   res->add(qr);
 
   strcpy(buf,"ccc.dat");
   qr=new QueryResult(buf,pid3);
   res->add(qr);
   
   PeerQueryResultMessage * qrm=new PeerQueryResultMessage(pid3,pid1,32,res);
   ser=pms->serialize(qrm);
   
   printf("[%s]\n",ser);
   
   PeerQueryResultMessage * qrm2=(PeerQueryResultMessage*)pms->deserialize(ser);
   printf("Message:\nTo: %s\nFrom: %s\n",qrm2->getToPeer()->toString(),qrm2->getFromPeer()->toString());
   
   List * l=qrm2->getResults();
   
   for (int i=0; i<l->getCount(); i++)
   {
      QueryResult * qr=(QueryResult*)l->getItem(i);
      printf("  Peer: %s    Result: %s\n",qr->getPeer()->toString(),qr->getData());
   }

   PeerID * pid4=new IPPeerID(130,207,104,100,2000);
   
   QueryProcessor * qp=cm->newQueryProcessor(pid4);

   ContentIndex * ci=qp->getContentIndex("mainindex");
   
   PeerIndexMessage * pim=new PeerIndexMessage(pid3,pid4,ci);
   
   ser=pms->serialize(pim);
   
   printf("[%s]\n",ser);

   PeerIndexMessage * pim2=(PeerIndexMessage*)pms->deserialize(ser);

   ser=pms->serialize(pim2);

   printf("[%s]\n",ser);

   PeerIndexRemoveMessage * pirm=new PeerIndexRemoveMessage(pid3,pid4);
   
   ser=pms->serialize(pirm);
   
   printf("[%s]\n",ser);

   PeerIndexRemoveMessage * pirm2=(PeerIndexRemoveMessage*)pms->deserialize(ser);

   ser=pms->serialize(pirm2);

   printf("[%s]\n",ser);

   return 0;
}

*/
