/**
 * ODIN overlay management middleware toolkit
 * 
 * ODIN Group, College of Computing, Georgia Institute of Technology
 *
 * For more information:
 * http://www.cc.gatech.edu/~cooperb/odin/
 * odin.gatech.edu
 *
 * You are free to use, modify and redistribute this software. 
 * However, we make no warranty for this software, and by downloading it
 * you agree not to hold use liable for anything. Do not under any
 * circumstances use this software to share copyrighted content or do
 * anything illegal. This software was designed for academic research
 * purposes and not for illegal activities. We are not responsible for
 * anything illegal you do and by downloading this software you agree
 * not to break any laws. Enjoy!
 *
 * Brian F. Cooper
 * November 21, 2005
 */


#ifndef PEERMESSAGES_H
#define PEERMESSAGES_H

#include "peer/Message.h"
#include "QueryProcessor.h"
#include "auxil/MessageSerializer.h"

const int PeerQueryMessageType=200;
const int PeerQueryResultMessageType=201;
const int PeerIndexMessageType=202;
const int PeerIndexRemoveMessageType=203;

const char PeerQueryMessageMagic[] = "%%PeerQueryMessage%%";
const char PeerQueryResultMessageMagic[] = "%%PeerQueryResultMessage%%";
const char PeerIndexMessageMagic[] = "%%PeerIndexMessage%%";
const char PeerIndexRemoveMessageMagic[] = "%%PeerIndexRemoveMessage%%";

/**
 * A search message sent between peers.
 */
class PeerQueryMessage : public Message
{
  public:
   /**
    * Create a search message. Each search has a unique id (_messageid) that is retained as the message is forwarded.
    * _searchingpeer and _query are cloned so you can delete them.
    */
   PeerQueryMessage(PeerID * _to, PeerID * _from, char * _query, PeerID * _searchingpeer, int _messageid, int _ttl, int _resultcount);

   virtual ~PeerQueryMessage();
   
   /**
    * Get the query. Don't delete this string.
    */
   char * getQuery();
   
   /**
    * Get the id of the original searching peer.
    */
   PeerID * getSearchingPeerID();
   
   /**
    * Get the id of this particular search.
    */
   int getMessageID();
   
   /**
    * Get the time-to-live of this message.
    */
   int getTTL();
   
   /**
    * Get the target result count, or -1 if none.
    */
   int getResultCount();

   char *getVisString();
   
  private:
   char * query;
   PeerID * searchingpeer;
   int messageid;
   int ttl;
   int resultcount;
};

/**
 * The results of a search.
 */
class PeerQueryResultMessage : public Message
{
  public:
   /**
    * Create a result message. The list of results will be kept (e.g. not cloned) and deleted when the message is deleted.
    */
   PeerQueryResultMessage(PeerID * _to, PeerID * _from, int _messageid, List * _results);
   
   virtual ~PeerQueryResultMessage();
   
   /**
    * Get the results, as a List of QueryResult objects.
    */
   List * getResults();
   
   /**
    * Get the results, and remove a reference to the list from this message object. Then, when the message
    * gets deleted, the list won't be deleted. This allows you to pass the results list to another object
    * without having to copy it. But future calls to extractResults() or getResults() on this object will return NULL.
    */
   List * extractResults();

   /**
    * Get the id of the search message that these are results for.
    */
   int getMessageID();

   char *getVisString();

  private:
   List * results;
   int messageid;
};

/**
 * A message to send a copy of an index, over an index link, to another peer.
 */
class PeerIndexMessage : public Message
{
  public:
   /**
    * Create a PeerIndexMessage. The _index is kept, and not cloned, so don't delete it after giving it to this object.
    */
   PeerIndexMessage(PeerID * _to, PeerID * _from, ContentIndex * _index);
   ~PeerIndexMessage();
   
   /**
    * Get the index. Do not delete this index.
    */
   ContentIndex * getIndex();
};

/**
 * A message to invalidate a replicated index.
 */
class PeerIndexRemoveMessage : public Message
{
  public:
   /**
    * Invalidate all indexes from the sending peer.
    */
   PeerIndexRemoveMessage(PeerID * _to, PeerID * _from);
   
   /**
    * Invalidate only the named index.
    */
   PeerIndexRemoveMessage(PeerID * _to, PeerID * _from, char * _indexname);

   /**
    * Get the name of the index to invalidate, or NULL if we want to invalidate all of them.
    */
   char * getIndexName();
};

/**
 * A SingleSerializer for serializing peer messages. If you are going to use the peer messages defined here, you should
 * register an instance of this serializer with the MessageSerializer for the system (e.g. the PeerExecutive's message serializer). Also,
 * the message serializer given in the constructor should be able to serialize ContentIndex objects.
 */
class PeerMessageSerializer : public SingleSerializer
{
  public:
   /**
    * Create a PeerMessageSerializer. The MessageSerializer given in this constructor should be able to serialize ContentIndex objects.
    */
   PeerMessageSerializer(MessageSerializer * _serializer);
   
   /**
    * Returns 1 if this instance of SingleSerializer knows how to serialize the given message. Otherwise, returns 0.
    */
   virtual int canSerialize(Message * m);
   
   /**
    * Returns 1 if this instance of SingleSerializer knows how to deserialize the given message. Otherwise, returns 0.
    */
   virtual int canDeserialize(char * buf);
   
   /**
    * Get a buffer containing the serialized version of the message. 
    * Delete the buffer when you are done with it.
    */
   virtual char * serialize(Message * m);
   
   /**
    * Deserialize a message buffer into a message. 
    */
   virtual Message * deserialize(char * buf);   

  private:
   MessageSerializer * serializer;
};

#endif
