/**
 * ODIN overlay management middleware toolkit
 * 
 * ODIN Group, College of Computing, Georgia Institute of Technology
 *
 * For more information:
 * http://www.cc.gatech.edu/~cooperb/odin/
 * odin.gatech.edu
 *
 * You are free to use, modify and redistribute this software. 
 * However, we make no warranty for this software, and by downloading it
 * you agree not to hold use liable for anything. Do not under any
 * circumstances use this software to share copyrighted content or do
 * anything illegal. This software was designed for academic research
 * purposes and not for illegal activities. We are not responsible for
 * anything illegal you do and by downloading this software you agree
 * not to break any laws. Enjoy!
 *
 * Brian F. Cooper
 * November 21, 2005
 */


#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include "peer/Runtime.h"
#include "peer/PeerLogic.h"
#include "peer/Logger.h"
#include "peer/TopologyManager.h"
#include "kernel/DelayMap.h"
#include "kernel/NetworkEmulatorExecutive.h"
#include "kernel/PeerExecutive.h"
#include "peer/IPPeerID.h"
#include "auxil/MessageSerializer.h"
#include "utils/Hashtable.h"
#include "utils/StringBuffer.h"
#include "common/Hostcatcher.h"
#include "kernel/TraceWorkloadExecutive.h"
#include "RandomWalkPeerLogicFactory.h"
#include "MinDegreeTopologyManager.h"
#include "SquareRootTopologyManager.h"
#include "MinSearchIndexDegreeTopologyManager.h"
#include "ContentMap.h"
#include "ContentMapQueryProcessor.h"
#include "BasicPeerLogic.h"
#include "common/ExtensiblePeerLogicFactory.h"
#include "common/ExtensibleTopologyManagerFactory.h"

const char * RequiredProperties[] = 
{
   "hostcatcheraddress",
   "tracefile",
   "neighborcheckinterval",
   "contentmapfile",
   "ttl",
   "resultcount",
   "peerlogictype",
   "topologymanagertype",
   "seenbeforesize",
};

const char * PropertyDescriptions[] =
{
   "The PeerID and port of the hostcatcher",
   "The file containing the trace that will drive the simulation",
   "The interval in ticks (ms) between checks of a peers degree and whether it needs to add or drop peers",
   "The file containing the content map for the simulation",
   "The TTL of walk messages",
   "The number of results that each search is looking for",
   "The kind of peer logic to create",
   "The kind of topology manager to create",
   "The number of messages in each peer's list of previously seen messages"
};

int RequiredPropertiesCount=9;

void usageMessage(char * arg0)
{
   printf("Usage: %s [-v n] [-r n] [-comment \"comment\"] property-file runid\n",arg0);
   printf("Options: \n");
   printf("  -comment \"comment\": make a comment about this run\n");
   printf("  -v n: use logging level n (0 is silent, 4 is most verbose)\n");
   printf("  -r n: use random seed n\n");
}

int main(int argc, char ** argv)
{
   unsigned int seed=time(NULL);
   
   int argindex=1;
   int debuglevel=LoggerUrgent;

   //parse command line options
   while ( (argindex<argc) && (argv[argindex][0]=='-') )
   {
      if (strcmp(argv[argindex],"-v")==0)
      {
	 argindex++;
	 debuglevel=atoi(argv[argindex]);
	 argindex++;
      }
      else if (strcmp(argv[argindex],"-r")==0)
      {
	 argindex++;
	 seed=atoi(argv[argindex]);
	 argindex++;
      }
      else if (strcmp(argv[argindex],"-comment")==0)
      {
	 argindex+=2;
      }
      else
      {
	 usageMessage(argv[0]);
	 return 0;
      }
   }

   if (argindex+2!=argc)
   {
      usageMessage(argv[0]);
      return 0;
   }

   srandom(seed);

   char * propertyfile=argv[argindex];
   argindex++;

   int runid=atoi(argv[argindex]);
   argindex++;
   
   // Load properties
   Properties * properties=new Properties(propertyfile);

   //check for required properties
   for (int i=0; i<RequiredPropertiesCount; i++)
   {
      if (properties->getProperty(RequiredProperties[i])==NULL)
      {
	 printf("Missing property: %s (%s)\n",RequiredProperties[i],PropertyDescriptions[i]);
	 return 0;
      }
   }

   // Set up simulation
   Supervisor * sup=new Supervisor();

   Measurements * meas=new Measurements(argc,argv,runid,1);
   sup->setMeasurements(meas);

   Logger * logger=new Logger(debuglevel);
   sup->setLogger(logger);
   
   ContentMap * contentmap=new ContentMap(properties->getProperty("contentmapfile"),NULL);

   //RandomWalkPeerLogicFactory *rwplf=new RandomWalkPeerLogicFactory(contentmap);
   ExtensiblePeerLogicFactory * eplf=new ExtensiblePeerLogicFactory();
   eplf->registerFactory(new HostcatcherPeerLogicFactory());
   eplf->registerFactory(new BasicPeerLogicFactory(contentmap));
   eplf->registerFactory(new RandomWalkPeerLogicFactory(contentmap,logger));

   ExtensibleTopologyManagerFactory * etmf=new ExtensibleTopologyManagerFactory();
   etmf->registerFactory(new HostcatcherTopologyManagerFactory());
   etmf->registerFactory(new MinDegreeTopologyManagerFactory(logger));
   etmf->registerFactory(new SquareRootTopologyManagerFactory(logger));
   etmf->registerFactory(new MinSearchIndexDegreeTopologyManagerFactory(logger));
   //MinDegreeTopologyManagerFactory * mdtmf=new MinDegreeTopologyManagerFactory(logger);
   //RandomWalkTopologyManagerFactory * rwtmf=new RandomWalkTopologyManagerFactory(mdtmf);
   

   DelayMap * dm=new DelayMap(20);

   NetworkEmulatorExecutive * nee=new NetworkEmulatorExecutive(sup,dm);
   PeerExecutive * pe=new PeerExecutive(sup,eplf,etmf,nee);
   TraceWorkloadExecutive *twe=new TraceWorkloadExecutive(properties->getProperty("tracefile"),sup,pe,NULL,properties);

   MessageSerializer * mserializer=new MessageSerializer();
   mserializer->registerSingleSerializer(new HostcatcherMessageSerializer(mserializer));
   mserializer->registerSingleSerializer(new PeerMessageSerializer(mserializer));
   mserializer->registerObjectSerializer(new ContentMapContentIndexSerializer(contentmap,mserializer));
   pe->setMessageSerializer(mserializer);

   sup->addExecutive(nee,1);
   sup->addExecutive(pe,1);
   sup->addExecutive(twe,1);
   
   Schema * messagestatistics=new Schema("MessageStatistics");
   messagestatistics->addField("Peer",TypeString);
   messagestatistics->addField("MessageType",TypeInt);
   messagestatistics->addField("MessagesSent",TypeInt);
   messagestatistics->addField("BytesSent",TypeLongLong);
   meas->newSchema(messagestatistics);

   Schema * degreestatistics=new Schema("DegreeStatistics");
   degreestatistics->addField("Peer",TypeString);
   degreestatistics->addField("SearchConnections",TypeInt);
   degreestatistics->addField("IndexConnections",TypeInt);
   meas->newSchema(degreestatistics);

   Schema * squarerootstatistics=new Schema("SquareRootStatistics");
   squarerootstatistics->addField("Peer",TypeString);
   squarerootstatistics->addField("TotalQueries",TypeInt);
   squarerootstatistics->addField("MatchedQueries",TypeInt);
   squarerootstatistics->addField("IdealDegree",TypeInt);
   meas->newSchema(squarerootstatistics);

   meas->printRun();
   meas->printSchema("MessageStatistics");
   meas->printSchema("DegreeStatistics");
   meas->printSchema("SquareRootStatistics");

   logger->message(LoggerInfo,"Starting simulation...");

   sup->run();

   logger->message(LoggerInfo,"Simulation finished.");

   //clean up
   delete messagestatistics;
   delete degreestatistics;
   delete squarerootstatistics;

   eplf->deleteAllFactories();
   delete eplf;
   etmf->deleteAllFactories();
   delete etmf;

   delete pe;
   delete nee;
   delete twe;
   delete dm;
   delete meas;
   delete contentmap;
   delete sup;
   delete properties;
   delete logger;
   mserializer->deleteAllSerializers();
   delete mserializer;

   return 0;
}
